/*
------------------------------------------------------------
	window.h - Generic Rendering Window
------------------------------------------------------------
 *  David Olofson, 2001
 * This code is released under the terms of the GPL.
 */

#include "window.h"
#include "gfxengine.h"
#include "sofont.h"

#define	SELECT	if(selected != this) select();

window_t *window_t::selected = NULL;

window_t::window_t()
{
	engine = NULL;
	rect.x = rect.y = 0;
	rect.w = 320;
	rect.h = 240;
	vx = vy = 0;
	gvx = gvy = 0;
	__gvx = __gvy = 0;
	vw = vh = 512;
	fgcolor = bgcolor = 0;
	selected = 0;
	xsc = ysc = 1;
}


window_t::~window_t()
{
	if(selected == this)
		selected = NULL;
}


void window_t::init(gfxengine_t *e)
{
	engine = e;
	xsc = e->xsc;
	ysc = e->ysc;
}


void window_t::place(int left, int top,
		int sizex, int sizey,
		int vsizex, int vsizey)
{
	rect.x = left * xsc;
	rect.y = top * ysc;
	rect.w = sizex * xsc;
	rect.h = sizey * ysc;
	vx = 0;
	vy = 0;
	gvx = 0;
	gvy = 0;
	__gvx = 0;
	__gvy = 0;

	if(vsizex < 0)
		vw = rect.w;
	else
		vw = vsizex * xsc;

	if(vsizey < 0)
		vh = rect.h;
	else
		vh = vsizey * ysc;
}


void window_t::select()
{
	if(!engine)
		return;

	SDL_Rect r = rect;
	selected = this;
	SDL_SetClipRect(engine->surface(), &r);
}


void window_t::invalidate(SDL_Rect * r = NULL)
{
	if(!engine)
		return;
	if(!engine->screen)
		return;

	if(!r)
		engine->invalidate(&rect);
	else
	{
		/* Translate to screen coordinates */
		SDL_Rect dr = *r;
		dr.x *= xsc;
		dr.y *= ysc;
		dr.w *= xsc;
		dr.h *= ysc;
		dr.x += rect.x;
		dr.y += rect.y;

		/* Clip to window (stolen from SDL_surface.c) */
		int Amin, Amax, Bmin, Bmax;

		/* Horizontal intersection */
		Amin = dr.x;
		Amax = Amin + dr.w;
		Bmin = rect.x;
		Bmax = Bmin + rect.w;
		if(Bmin > Amin)
			Amin = Bmin;
		dr.x = Amin;
		if(Bmax < Amax)
			Amax = Bmax;
		dr.w = Amax - Amin > 0 ? Amax - Amin : 0;

		/* Vertical intersection */
		Amin = dr.y;
		Amax = Amin + dr.h;
		Bmin = rect.y;
		Bmax = Bmin + rect.h;
		if(Bmin > Amin)
			Amin = Bmin;
		dr.y = Amin;
		if(Bmax < Amax)
			Amax = Bmax;
		dr.h = Amax - Amin > 0 ? Amax - Amin : 0;

		if(dr.w && dr.h)
			engine->invalidate(&dr);
	}
}


#if 0
void window_t::set_position(int vposx, int vposy)
{
	vx = vposx;
	vy = vposy;
	if(engine->output() != this)
	{
		gvx = vx;
		gvy = vy;
	}
}


void window_t::force_position(int vposx, int vposy)
{
	set_position(vposx, vposy);
	engine->force_scroll();
}
#endif


/*---------------------------------------------------------------
	Rendering API
---------------------------------------------------------------*/

Uint32 window_t::alloc_color(int r, int g, int b)
{
	if(!engine)
		return 0;

	return SDL_MapRGB(engine->surface()->format, r>>8, g>>8, b>>8);
}

void window_t::foreground(Uint32 color)
{
	fgcolor = color;
}

void window_t::background(Uint32 color)
{
	bgcolor = color;
}


void window_t::font(int fnt)
{
	_font = fnt;
}


void window_t::string(int x, int y, const char *txt)
{
	string_fxp(PIXEL2CS(x), PIXEL2CS(y), txt);
}


void window_t::center(int y, const char *txt)
{
	center_fxp(PIXEL2CS(y), txt);
}


void window_t::string_fxp(int x, int y, const char *txt)
{
	if(!engine)
		return;

	x = CS2PIXEL(x*xsc);
	y = CS2PIXEL(y*ysc);
	SoFont *f = engine->get_font(_font);
	if(!f)
		return;
	SELECT
	x += rect.x - gvx;
	y += rect.y - gvy;
	f->PutString(engine->surface(), x, y, txt);
}


void window_t::center_fxp(int y, const char *txt)
{
	y = CS2PIXEL(y*ysc);

	if(!engine)
		return;

	SoFont *f = engine->get_font(_font);
	if(!f)
		return;
	SELECT
	int x = (rect.w - f->TextWidth(txt)) / 2;
	x += rect.x - gvx;
	y += rect.y - gvy;
	f->PutString(engine->surface(), x, y, txt);
}


int window_t::textwidth(const char *txt)
{
	if(!engine)
		return strlen(txt);

	SoFont *f = engine->get_font(_font);
	if(!f)
		return strlen(txt);

	return f->TextWidth(txt) / xsc;
}


void window_t::clear()
{
	if(!engine)
		return;
	SELECT
	SDL_Rect r = rect;
	SDL_FillRect(engine->surface(), &r, bgcolor);
}


void window_t::point(int x, int y)
{
	x *= xsc;
	y *= ysc;

	if(!engine)
		return;
	SELECT
	/* Quick hack; slow */
	SDL_Rect r;
	r.x = rect.x + x - gvx;
	r.y = rect.y + y - gvy;
	r.w = xsc;
	r.h = ysc;
	SDL_FillRect(engine->surface(), &r, fgcolor);
}


void window_t::rectangle(int x, int y, int w, int h)
{
	x *= xsc;
	y *= ysc;
	w *= xsc;
	h *= ysc;

	if(!engine)
		return;
	SELECT
	SDL_Rect r;
	r.x = rect.x + x - gvx;
	r.y = rect.y + y - gvy;
	r.w = w;
	r.h = ysc;
	SDL_FillRect(engine->surface(), &r, fgcolor);

	r.y = rect.y + y+h-ysc - gvy;
	SDL_FillRect(engine->surface(), &r, fgcolor);

	r.x = rect.x + x - gvx;
	r.y = rect.y + y+ysc - gvy;
	r.w = xsc;
	r.h = h-2*ysc;
	SDL_FillRect(engine->surface(), &r, fgcolor);

	r.x = rect.x + x+w-xsc - gvx;
	SDL_FillRect(engine->surface(), &r, fgcolor);
}


void window_t::fillrect(int x, int y, int w, int h)
{
	x *= xsc;
	y *= ysc;
	w *= xsc;
	h *= ysc;

	if(!engine)
		return;

	SELECT
	SDL_Rect r;
	r.x = rect.x + x - gvx;
	r.y = rect.y + y - gvy;
	r.w = w;
	r.h = h;
	SDL_FillRect(engine->surface(), &r, fgcolor);
}


void window_t::fillrect_fxp(int x, int y, int w, int h)
{
	x = CS2PIXEL(x*xsc);
	y = CS2PIXEL(y*ysc);
	w = CS2PIXEL(w*xsc);
	h = CS2PIXEL(h*ysc);

	if(!engine)
		return;

	SELECT
	SDL_Rect r;
	r.x = rect.x + x - gvx;
	r.y = rect.y + y - gvy;
	r.w = w;
	r.h = h;
	SDL_FillRect(engine->surface(), &r, fgcolor);
}


void window_t::sprite(int x, int y, int bank, int frame, int inval = 1)
{
	sprite_fxp(PIXEL2CS(x), PIXEL2CS(y), bank, frame, inval);
}


void window_t::sprite_fxp(int x, int y, int bank, int frame, int inval = 1)
{
	if(!engine)
		return;
	x = CS2PIXEL(x*xsc);
	y = CS2PIXEL(y*ysc);
	SELECT
	SDL_Rect dest_rect;
	SDL_Surface *img;
	img = engine->get_sprite(bank, frame);
	if(!img)
		return;

	dest_rect.x = rect.x + x - gvx;
	dest_rect.y = rect.y + y - gvy;
	SDL_BlitSurface(img, NULL, engine->surface(), &dest_rect);

	if(inval && !engine->autoinvalidate)
	{
		dest_rect.w = img->w;
		dest_rect.h = img->h;
		engine->invalidate(&dest_rect);
	}
}
