/*
------------------------------------------------------------
   Kobo Deluxe - An enhanced SDL port of XKobo
------------------------------------------------------------
 * Copyright (C) 2001, David Olofson
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <string.h>
#include <stdlib.h>

#include "prefs.h"
#include "config.h"
#include "gfxengine.h"

prefs_t::prefs_t()
{
	set_defaults();
}


prefs_t::~prefs_t()
{
}


void prefs_t::set_defaults()
{
	changed = 0;

	use_joystick = 0;

	fullscreen = 0;
	videodriver = GFX_DRIVER_SDL2D;
	depth = 0;	//use default/current
	buffermode = GFX_BUFFER_SEMITRIPLE;
//	size = 1;
	width = 320;
	height = 240;
	internalres = 1;
	scalemode = 2;	//Interpolation with oversampling

	wait_msec = WAIT_MSEC;
	filter = 1;	//use "motion filtering"

	cheat_mode = 0;
	indicator = 0;

	sound = 1;
	samplerate = 44100;
	latency = 50;
	volume = 100;
	reverb = 100;
	threshold = 200;
	release = 50;
	use_ext_music = 0;
	use_oss = 0;

	bgm_indexfile[0] = 0;
	dir[0] = 0;
	gfxdir[0] = 0;
	scoredir[0] = 0;
}


int prefs_t::parse(int argc, char *argv[])
{
	int i;
	DBG(for (i = 1; i < argc; i++)
		fprintf(stderr, "'%s'\n", argv[i]));

	for (i = 1; i < argc; i++)
	{
		if(argv[i][0] == '#')
			continue;

		int yesno = 1;
		char *a = argv[i];
		if(a[0] == '-' || a[0] == '/')
			++a;

		if(!strncmp("no", a, 2))
		{
			yesno = 0;
			a += 2;
		}

		if(!strcmp("fullscreen", a))
			fullscreen = yesno;
		else if(!strcmp("filter", a))
			filter = yesno;
		else if(!strcmp("cheat", a))
			cheat_mode = yesno;
		else if(!strcmp("indicator", a))
			indicator = yesno;
		else if(!strcmp("sound", a))
			sound = yesno;
		else if(!strcmp("oss", a))
			use_oss = yesno;
		else if(!strcmp("joystick", a))
			use_joystick = yesno;
		else if(!strcmp("frame", a))
			continue;

		else if(yesno == 0)	// The rest don't have "no*" variants!
			return -1;

		else if(!strcmp("showcfg", a))
			continue;
		else if(!strcmp("hiscores", a))
			continue;
		else if(!strcmp("highscores", a))
			continue;
		else if(!strcmp("override", a))
			continue;
		else if(!strcmp("debug", a))
			continue;
		else if(!strcmp("fps", a))
			continue;

		else
		{
			/*
			 * Parse args with one parameter
			 */
			if(argc <= i+1)
			{
				fprintf(stderr, "ERROR: Argument '%s'"
						" needs an parameter!\n",
						a);
				return -2;
			}

			if(!strcmp("files", a))
				strncpy(dir, argv[++i], 256);
			else if(!strcmp("gfx", a))
				strncpy(gfxdir, argv[++i], 256);
			else if(!strcmp("scores", a))
				strncpy(scoredir, argv[++i], 256);
			else if(!strcmp("depth", a))
				depth = atoi(argv[++i]);
			else if(!strcmp("buffer", a))
				buffermode = atoi(argv[++i]);
			else if(!strcmp("size", a))
			{
				/* For pre 20011007 compatibility */
				int size = atoi(argv[++i]);
				width = 320 * size;
				height = 240 * size;
			}
			else if(!strcmp("width", a))
				width = atoi(argv[++i]);
			else if(!strcmp("height", a))
				height = atoi(argv[++i]);
			else if(!strcmp("videodriver", a))
				videodriver = atoi(argv[++i]);
			else if(!strcmp("internalres", a))
				internalres = atoi(argv[++i]);
			else if(!strcmp("scalemode", a))
				scalemode = atoi(argv[++i]);
			else if(!strcmp("wait", a))
				wait_msec = atoi(argv[++i]);
			else if(!strcmp("samplerate", a))
				samplerate = atoi(argv[++i]);
			else if(!strcmp("latency", a))
				latency = atoi(argv[++i]);
			else if(!strcmp("bgm", a))
				strncpy(bgm_indexfile, argv[++i], 256);
			else if(!strcmp("vol", a))
				volume = atoi(argv[++i]);
			else if(!strcmp("reverb", a))
				reverb = atoi(argv[++i]);
			else if(!strcmp("threshold", a))
				threshold = atoi(argv[++i]);
			else if(!strcmp("release", a))
				release = atoi(argv[++i]);
			else
			{
				fprintf(stderr, "ERROR: Unknown argument"
						" '%s'!\n", a);
				return -3;
			}
		}
	}
	return 0;
}


enum scanstate_t
{
	SS_BLANK,
	SS_QUOTE,
	SS_WORD,
	SS_COMMENT,
};

int prefs_t::read_config(char ***cv, FILE *f)
{
	*cv = (char **)calloc(MAX_CONFIG_LINES, sizeof(char *));
	if(!*cv)
		return -1;

	if(!f)
	{
		free((*cv));
		(*cv) = NULL;
		return -2;
	}

	fseek(f, 0, SEEK_END);
	int len = ftell(f);
	(*cv)[0] = (char *)malloc(len+1);
	if(!(*cv)[0])
	{
		rewind(f);
		free(*cv);
		*cv = NULL;
		return -3;
	}
	rewind(f);
	if(fread((*cv)[0], len, 1, f) < 0)
	{
		free(*cv[0]);
		free(*cv);
		*cv = NULL;
		return -4;
	}
	(*cv)[0][len] = 0;

	char *v = (*cv)[0];
	scanstate_t state = SS_BLANK;
	int arg = 1;
	int start_arg = 0;
	while(*v)
	{
		switch(state)
		{
		  case SS_BLANK:
			if(*v == '"')
			{
				start_arg = 1;
				state = SS_QUOTE;
			}
			else if(*v == '#')
				state = SS_COMMENT;
			else if(*v > ' ')
			{
				start_arg = 1;
				state = SS_WORD;
			}
			break;
		  case SS_QUOTE:
			if(*v == '"')
			{
				*v = 0;
				state = SS_BLANK;
			}
			break;
		  case SS_WORD:
			if(*v <= ' ')
			{
				*v = 0;
				state = SS_BLANK;
			}
			break;
		  case SS_COMMENT:
			if((*v == 10) || (*v == 13))
				state = SS_BLANK;
			break;
		}
		if(start_arg)
		{
			if(arg < MAX_CONFIG_LINES)
			{
				if(state == SS_QUOTE)
					(*cv)[arg++] = v+1;
				else
					(*cv)[arg++] = v;
			}
			start_arg = 0;
		}
		++v;
	}
	rewind(f);
	return arg;
}


int prefs_t::read(FILE *f)
{
	int res = -10;
	int cc;
	char **cv;
	cc = read_config(&cv, f);
	if(cc > 0)
	{
		res = parse(cc, cv);

		if(cv)
			free(cv[0]);
		free(cv);
	}
	return res;
}


void prefs_t::print_switch(FILE *f, const char *base, int flag)
{
	fprintf(f, flag ? "%s\n" : "no%s\n", base);
}


int prefs_t::write(FILE *f)
{
	fprintf(f, "#---------------------------------------\n");
	fprintf(f, "# Kobo Deluxe "VERSION" Configuration File\n");
	fprintf(f, "#---------------------------------------\n");
	fprintf(f, "# Switches - [no]<switch>\n");
	print_switch(f, "fullscreen", fullscreen);
	print_switch(f, "filter", filter);
	print_switch(f, "cheat", cheat_mode);
	print_switch(f, "indicator", indicator);
	print_switch(f, "sound", sound);
#ifdef OSS_SUPPORT
	print_switch(f, "oss", use_oss);
#endif
	print_switch(f, "joystick", use_joystick);

	fprintf(f, "\n# Values - <key> [<value>|\"<string>\"]\n");
	fprintf(f, "bgm \"%s\"\n", bgm_indexfile);
	fprintf(f, "files \"%s\"\n", dir);
	fprintf(f, "gfx \"%s\"\n", gfxdir);
	fprintf(f, "scores \"%s\"\n", scoredir);
	fprintf(f, "depth %d\n", depth);
	fprintf(f, "buffer %d\n", buffermode);
	fprintf(f, "wait %d\n", wait_msec);
	fprintf(f, "samplerate %d\n", samplerate);
	fprintf(f, "latency %d\n", latency);
	fprintf(f, "vol %d\n", volume);
	fprintf(f, "reverb %d\n", reverb);
	fprintf(f, "threshold %d\n", threshold);
	fprintf(f, "release %d\n", release);
//	fprintf(f, "size %d\n", size);
	fprintf(f, "width %d\n", width);
	fprintf(f, "height %d\n", height);
	fprintf(f, "videodriver %d\n", videodriver);
	fprintf(f, "internalres %d\n", internalres);
	fprintf(f, "scalemode %d\n", scalemode);

	return 0;
}
